# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the LIMEExplainer for computing explanations on black box models using LIME."""

import numpy as np
from interpret_community._internal.raw_explain.raw_explain_utils import (
    get_datamapper_and_transformed_data, transform_with_datamapper)
from interpret_community.common.aggregate import add_explain_global_method
from interpret_community.common.blackbox_explainer import (
    BlackBoxExplainer, add_prepare_function_and_summary_method,
    init_blackbox_decorator)
from interpret_community.common.constants import (Defaults, ExplainParams,
                                                  ExplainType,
                                                  ExplanationParams, Extension,
                                                  ModelTask)
from interpret_community.common.model_wrapper import _wrap_model
from interpret_community.common.progress import get_tqdm
from interpret_community.dataset.decorator import (init_tabular_decorator,
                                                   tabular_decorator)
from interpret_community.explanation.explanation import (
    _create_local_explanation, _create_raw_feats_local_explanation,
    _get_raw_explainer_create_explanation_kwargs)

# Soft dependency for LIME
try:
    lime_import_failed = False
    from lime.lime_tabular import LimeTabularExplainer
except ImportError:
    lime_import_failed = True

import warnings

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', 'Starting from version 2.2.1', UserWarning)


@add_prepare_function_and_summary_method
@add_explain_global_method
class LIMEExplainer(BlackBoxExplainer):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.BLACKBOX

    """Defines the LIME Explainer for explaining black box models or functions.

    :param model: The model to explain or function if is_function is True.
        A model that implements sklearn.predict or sklearn.predict_proba or function that accepts a 2d
        ndarray.
    :type model: object
    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param is_function: Default set to false, set to True if passing function instead of model.
    :type is_function: bool
    :param explain_subset: List of feature indices. If specified, only selects a subset of the
        features in the evaluation dataset for explanation. The subset can be the top-k features
        from the model summary.
    :type explain_subset: list[int]
    :param nclusters: Number of means to use for approximation. A dataset is summarized with nclusters mean
        samples weighted by the number of data points they each represent. When the number of initialization
        examples is larger than (10 x nclusters), those examples will be summarized with k-means where
        k = nclusters.
    :type nclusters: int
    :param features: A list of feature names.
    :type features: list[str]
    :param classes: Class names as a list of strings. The order of the class names should match
        that of the model output.  Only required if explaining classifier.
    :type classes: list[str]
    :param verbose: If true, uses verbose logging in LIME.
    :type verbose: bool
    :param categorical_features: Categorical feature names or indexes.
        If names are passed, they will be converted into indexes first.
    :type categorical_features: Union[list[str], list[int]]
    :param show_progress: Default to 'True'.  Determines whether to display the explanation status bar
        when using LIMEExplainer.
    :type show_progress: bool
    :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
    transformer. When transformations are provided, explanations are of the features before the transformation.
    The format for list of transformations is same as the one here:
    https://github.com/scikit-learn-contrib/sklearn-pandas.

    If the user is using a transformation that is not in the list of sklearn.preprocessing transformations that
    we support then we cannot take a list of more than one column as input for the transformation.
    A user can use the following sklearn.preprocessing  transformations with a list of columns since these are
    already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
    MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer, RobustScaler,
    StandardScaler.

    Examples for transformations that work::

        [
            (["col1", "col2"], sklearn_one_hot_encoder),
            (["col3"], None) #col3 passes as is
        ]
        [
            (["col1"], my_own_transformer),
            (["col2"], my_own_transformer),
        ]

    Example of transformations that would raise an error since it cannot be interpreted as one to many::

        [
            (["col1", "col2"], my_own_transformer)
        ]

    This would not work since it is hard to make out whether my_own_transformer gives a many to many or one to
    many mapping when taking a sequence of columns.
    :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
    :param allow_all_transformations: Allow many to many and many to one transformations
    :type allow_all_transformations: bool
    :param model_task: Optional parameter to specify whether the model is a classification or regression model.
        In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
        has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
        outputs a 1 dimensional array.
    :type model_task: str
    """

    @init_tabular_decorator
    @init_blackbox_decorator
    def __init__(self, model, initialization_examples, is_function=False, explain_subset=None,
                 nclusters=10, features=None, classes=None, verbose=False,
                 categorical_features=[],  # noqa: B006
                 show_progress=True, transformations=None, allow_all_transformations=False,
                 model_task=ModelTask.Unknown, **kwargs):
        """Initialize the LIME Explainer.

        :param model: The model to explain or function if is_function is True.
            A model that implements sklearn.predict or sklearn.predict_proba or function that accepts a 2d
            ndarray.
        :type model: object
        :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param is_function: Default set to false, set to True if passing function instead of model.
        :type is_function: bool
        :param explain_subset: List of feature indices. If specified, only selects a subset of the
            features in the evaluation dataset for explanation. The subset can be the top-k features
            from the model summary.
        :type explain_subset: list[int]
        :param nclusters: Number of means to use for approximation. A dataset is summarized with nclusters mean
            samples weighted by the number of data points they each represent. When the number of initialization
            examples is larger than (10 x nclusters), those examples will be summarized with k-means where
            k = nclusters.
        :type nclusters: int
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :param verbose: If true, uses verbose logging in LIME.
        :type verbose: bool
        :param categorical_features: Categorical feature names or indexes.
            If names are passed, they will be converted into indexes first.
        :type categorical_features: Union[list[str], list[int]]
        :param show_progress: Default to 'True'.  Determines whether to display the explanation status bar
            when using LIMEExplainer.
        :type show_progress: bool
        :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
        transformer. When transformations are provided, explanations are of the features before the transformation.
        The format for list of transformations is same as the one here:
        https://github.com/scikit-learn-contrib/sklearn-pandas.

        If the user is using a transformation that is not in the list of sklearn.preprocessing transformations that
        we support then we cannot take a list of more than one column as input for the transformation.
        A user can use the following sklearn.preprocessing  transformations with a list of columns since these are
        already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
        MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer, RobustScaler,
        StandardScaler.

        Examples for transformations that work::

            [
                (["col1", "col2"], sklearn_one_hot_encoder),
                (["col3"], None) #col3 passes as is
            ]
            [
                (["col1"], my_own_transformer),
                (["col2"], my_own_transformer),
            ]

        Example of transformations that would raise an error since it cannot be interpreted as one to many::

            [
                (["col1", "col2"], my_own_transformer)
            ]

        This would not work since it is hard to make out whether my_own_transformer gives a many to many or one to
        many mapping when taking a sequence of columns.
        :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :param model_task: Optional parameter to specify whether the model is a classification or regression model.
            In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
            has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
            outputs a 1 dimensional array.
        :type model_task: str
        """
        if lime_import_failed:
            raise Exception("Could not import LIME, required for LIMEExplainer")
        self._datamapper = None
        if transformations is not None:
            self._datamapper, initialization_examples = get_datamapper_and_transformed_data(
                examples=initialization_examples, transformations=transformations,
                allow_all_transformations=allow_all_transformations)
        wrapped_model, eval_ml_domain = _wrap_model(model, initialization_examples, model_task, is_function)
        super(LIMEExplainer, self).__init__(wrapped_model, is_function=is_function,
                                            model_task=eval_ml_domain, **kwargs)
        self._logger.debug('Initializing LIMEExplainer')

        self._method = 'lime'
        self.initialization_examples = initialization_examples
        self.classification = False
        self.features = initialization_examples.get_features(features=features)
        self.classes = classes
        self.nclusters = nclusters
        self.explain_subset = explain_subset
        self.show_progress = show_progress
        self.transformations = transformations
        # If categorical_features is a list of string column names instead of indexes, make sure to convert to indexes
        if not all(isinstance(categorical_feature, int) for categorical_feature in categorical_features):
            categorical_features = initialization_examples.get_column_indexes(self.features, categorical_features)
        # Index the categorical string columns
        self._column_indexer = initialization_examples.string_index(columns=categorical_features)
        function, summary = self._prepare_function_and_summary(self.function, self.original_data_ref,
                                                               self.current_index_list, nclusters=nclusters,
                                                               explain_subset=explain_subset, **kwargs)
        if str(type(summary)).endswith(".DenseData'>"):
            summary = summary.data
        self._lime_feature_names = [str(i) for i in range(summary.shape[1])]
        result = function(summary[0].reshape((1, -1)))
        # If result is 2D array, this is classification scenario, otherwise regression
        if len(result.shape) == 2:
            self.classification = True
            mode = ExplainType.CLASSIFICATION
        elif len(result.shape) == 1:
            self.classification = False
            mode = ExplainType.REGRESSION
        else:
            raise Exception('Invalid function specified, does not conform to specifications on prediction')
        self.explainer = LimeTabularExplainer(summary, feature_names=self._lime_feature_names, class_names=classes,
                                              categorical_features=categorical_features, verbose=verbose,
                                              mode=mode, discretize_continuous=False)
        self.explainer.function = function
        if self.classes is None and self.classification:
            raise ValueError('LIME Explainer requires classes to be specified if using a classification model')
        if self.classes is not None and not self.classification:
            if self.model is None:
                error = 'Classes is specified but function was predict, not predict_proba.'
            else:
                error = 'Classes is specified but model does not define predict_proba, only predict.'
            raise ValueError(error)

    @tabular_decorator
    def explain_global(self, evaluation_examples, sampling_policy=None,
                       include_local=True, batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Explain the model globally by aggregating local explanations to global.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param sampling_policy: Optional policy for sampling the evaluation examples.  See documentation on
            SamplingPolicy for more information.
        :type sampling_policy: interpret_community.common.policy.SamplingPolicy
        :param include_local: Include the local explanations in the returned global explanation.
            If include_local is False, will stream the local explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: A model explanation object containing the global explanation.
        :rtype: GlobalExplanation
        """
        kwargs = {ExplainParams.METHOD: ExplainType.LIME,
                  ExplainParams.SAMPLING_POLICY: sampling_policy,
                  ExplainParams.INCLUDE_LOCAL: include_local,
                  ExplainParams.BATCH_SIZE: batch_size}

        if self.classification:
            kwargs[ExplanationParams.CLASSES] = self.classes
            kwargs[ExplainType.MODEL_TASK] = ExplainType.CLASSIFICATION
        else:
            kwargs[ExplainType.MODEL_TASK] = ExplainType.REGRESSION

        kwargs[ExplainParams.EVAL_DATA] = evaluation_examples.typed_dataset

        return self._explain_global(evaluation_examples, **kwargs)

    @tabular_decorator
    def explain_local(self, evaluation_examples):
        """Explain the function locally by using LIME.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: interpret_community.dataset.dataset_wrapper.DatasetWrapper
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :return: A model explanation object containing the local explanation.
        :rtype: LocalExplanation
        """
        if self._datamapper is not None:
            evaluation_examples = transform_with_datamapper(evaluation_examples, self._datamapper)

        if self._column_indexer:
            evaluation_examples.apply_indexer(self._column_indexer)

        # Compute subset info prior
        if self.explain_subset:
            evaluation_examples.take_subset(self.explain_subset)

        # sample the evaluation examples
        # note: the sampled data is also used by KNN
        if self.sampling_policy is not None and self.sampling_policy.allow_eval_sampling:
            sampling_method = self.sampling_policy.sampling_method
            max_dim_clustering = self.sampling_policy.max_dim_clustering
            evaluation_examples.sample(max_dim_clustering, sampling_method=sampling_method)
        features = self.features
        if self.explain_subset:
            features = [features[i] for i in self.explain_subset]
        kwargs = {ExplainParams.METHOD: ExplainType.LIME}
        kwargs[ExplainParams.FEATURES] = features
        kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features
        original_evaluation = evaluation_examples.original_dataset
        evaluation_examples = evaluation_examples.dataset
        if len(evaluation_examples.shape) == 1:
            evaluation_examples = evaluation_examples.reshape(1, -1)

        self._logger.debug('Running LIMEExplainer')
        if self.classification:
            kwargs[ExplanationParams.CLASSES] = self.classes
            kwargs[ExplainType.MODEL_TASK] = ExplainType.CLASSIFICATION
            num_classes = len(self.classes)
            labels = list(range(num_classes))
        else:
            kwargs[ExplainType.MODEL_TASK] = ExplainType.REGRESSION
            num_classes = 1
            labels = None
        lime_explanations = []

        tqdm = get_tqdm(self._logger, self.show_progress)

        if self.explain_subset:
            self.original_data_ref[0] = original_evaluation
            self.current_index_list.append(0)
            for ex_idx, example in tqdm(enumerate(evaluation_examples)):
                self.current_index_list[0] = ex_idx
                lime_explanations.append(self.explainer.explain_instance(example,
                                                                         self.explainer.function,
                                                                         labels=labels))
            self.current_index_list = [0]
        else:
            for _, example in tqdm(enumerate(evaluation_examples)):
                lime_explanations.append(self.explainer.explain_instance(example,
                                                                         self.explainer.function,
                                                                         labels=labels))
        if self.classification:
            lime_values = [None] * num_classes
            for lime_explanation in lime_explanations:
                for label in labels:
                    map_values = dict(lime_explanation.as_list(label=label))
                    if lime_values[label - 1] is None:
                        lime_values[label - 1] = [[map_values.get(feature, 0.0) for feature in
                                                   self._lime_feature_names]]
                    else:
                        lime_values[label - 1].append([map_values.get(feature, 0.0) for feature in
                                                       self._lime_feature_names])
        else:
            lime_values = None
            for lime_explanation in lime_explanations:
                map_values = dict(lime_explanation.as_list())
                if lime_values is None:
                    lime_values = [[map_values.get(feature, 0.0) for feature in self._lime_feature_names]]
                else:
                    lime_values.append([map_values.get(feature, 0.0) for feature in self._lime_feature_names])
        expected_values = None
        if self.model is not None:
            kwargs[ExplainParams.MODEL_TYPE] = str(type(self.model))
        else:
            kwargs[ExplainParams.MODEL_TYPE] = ExplainType.FUNCTION

        kwargs[ExplainParams.CLASSIFICATION] = self.classification
        kwargs[ExplainParams.LOCAL_IMPORTANCE_VALUES] = np.array(lime_values)
        kwargs[ExplainParams.EXPECTED_VALUES] = np.array(expected_values)
        kwargs[ExplainParams.EVAL_DATA] = original_evaluation

        explanation = _create_local_explanation(**kwargs)

        # if transformations have been passed, then return raw features explanation
        raw_kwargs = _get_raw_explainer_create_explanation_kwargs(kwargs=kwargs)
        return explanation if self._datamapper is None else _create_raw_feats_local_explanation(
            explanation, feature_maps=[self._datamapper.feature_map], features=self.features, **raw_kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from abc import ABC, abstractmethod

import numpy as np
from scipy.sparse import csr_matrix, eye
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (Binarizer, KernelCenterer, LabelEncoder,
                                   MaxAbsScaler, MinMaxScaler, Normalizer,
                                   OneHotEncoder, QuantileTransformer,
                                   RobustScaler, StandardScaler)

from ...common.constants import Scipy


def get_feature_mapper_for_pipeline(pipeline_obj):
    """Get FeatMapper object from a sklearn.pipeline.Pipeline object.

    :param pipeline_obj: pipeline object
    :type pipeline_obj: sklearn.pipeline.Pipeline
    :return: feat mapper for pipeline
    :rtype: PipelineFeatureMapper
    """
    """Get feat mapper for a pipeline, iterating over the transformers."""

    steps = []
    count = 0
    for _, transformer in pipeline_obj.steps:
        transformer_type = type(transformer)
        if transformer_type in encoders_to_mappers_dict:
            steps.append((str(count), encoders_to_mappers_dict[transformer_type](transformer)))
            count += 1
        elif transformer_type == Pipeline:
            steps.append((str(count), get_feature_mapper_for_pipeline(transformer)))
        else:
            steps.append((str(count), ManytoManyMapper(transformer)))

    return PipelineFeatureMapper(Pipeline(steps))


class FeatureMapper(ABC):
    """A class that supports both feature map from raw to engineered as well as a transform method."""

    @abstractmethod
    def __init__(self, transformer):
        """

        :param transformer: object that has a .transform method
        :type transformer: class that has a .transform method
        """
        self._transformer = transformer
        self._feature_map = None

    @property
    def transformer(self):
        return self._transformer

    @abstractmethod
    def transform(self, x):
        """

        :param x: input data
        :type x: pandas.DataFrame or numpy.array
        :rtype: transformed data
        :rtype: numpy.array or scipy.sparse matrix
        """
        pass

    @property
    def feature_map(self):
        """

        :return: feature map from raw to generated
        :rtype: numpy.array
        """
        if self._feature_map is None:
            raise ValueError("transform not called")
        return self._feature_map

    def set_feature_map(self, feature_map):
        """

        :param feature_map: feature map associated with the FeatureMapper
        :type feature_map: numpy.array
        """
        self._feature_map = feature_map

    def fit(self, x):
        """Dummy fit so that this can go in an sklearn.pipeline.Pipeline.

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: self
        :rtype: FeatureMapper
        """
        return self


class IdentityMapper(FeatureMapper):
    """FeatMapper for one to one mappings."""

    def __init__(self, transformer):
        """

        :param transformer: object that has a .transform method
        :type transformer: class that has a .transform method
        """
        super(IdentityMapper, self).__init__(transformer)

    def _build_feature_map(self, num_cols):
        """Build a feature map for one to one mappings from raw to generated.

        :param num_cols: number of columns in input.
        :type num_cols: int
        """
        self.set_feature_map(eye(num_cols, format=Scipy.CSR_FORMAT))

    def transform(self, x):
        """Transform input data.

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: transformed data
        :rtype: numpy.array
        """
        result = self.transformer.transform(x)
        if self._feature_map is None:
            self._build_feature_map(result.shape[1])

        return result


class PassThroughMapper(FeatureMapper):
    """FeatureMapper to use when only one column is the input."""

    def __init__(self, transformer):
        """

        :param transformer: object that has a .transform method
        :type transformer: class that has a .transform method
        """
        super(PassThroughMapper, self).__init__(transformer)

    def _build_feature_map(self, num_output_cols):
        """Build a feature map for mappings from raw to generated when the input column is just one.

        :param num_output_cols: number of output columns
        :type num_output_cols: [int]
        """
        self.set_feature_map(np.ones((1, num_output_cols)))

    def transform(self, x):
        """

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: transformed data
        :rytpe: numpy.array or scipy.sparse matrix
        """
        x_transformed = self.transformer.transform(x)
        if self._feature_map is None:
            self._build_feature_map(x_transformed.shape[1])

        return x_transformed


class PipelineFeatureMapper(FeatureMapper):
    """FeatureMapper for a sklearn pipeline of feat mappers"""

    def __init__(self, transformer):
        """

        :param transformer: object that has a .transform method
        :type transformer: class that has a .transform method
        """
        super(PipelineFeatureMapper, self).__init__(transformer)

    def _build_feature_map(self, feature_mappers):
        """Build a feature map for mappings from raw to generated for a pipeline of FeatMapper's.

        :param feature_mappers: list of feat mappers
        :type feature_mappers: [FeatureMapper]
        """

        curr_map = feature_mappers[0].feature_map
        for next_map in feature_mappers[1:]:
            curr_map = curr_map.dot(next_map.feature_map)

        self.set_feature_map(curr_map)

    def transform(self, x):
        """

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: transformed data
        :rtype: numpy.array or scipy.sparse matrix
        """
        ret = self.transformer.transform(x)
        if self._feature_map is None:
            self._build_feature_map([s[1] for s in self._transformer.steps])

        return ret


class OneHotEncoderMapper(FeatureMapper):
    """OneHotEncoder FeatureMapper"""
    def __init__(self, transformer):
        """Build a feature map for OneHotEncoder.

        :param transformer: object of type onehotencoder
        :type transformer: sklearn.preprocessing.OneHotEncoder
        """
        super(OneHotEncoderMapper, self).__init__(transformer)

    def _cat_len(self, cat, drop):
        """Internal method to compute the output length of OHE for categoricals.

        :param cat: The categorical features per column.
        :type cat: list
        :param drop: The drop parameter for OHE.  Can be 'None', 'first' or 'if_binary'.
        :type drop: str
        """
        cat_len = len(cat)
        if drop == 'first':
            return cat_len - 1
        elif drop == 'if_binary':
            if cat_len == 2:
                return 1
        return cat_len

    def _build_feature_map(self):
        """Build feature map when transformer is a one hot encoder."""
        drop = self.transformer.drop
        cat_lens = [self._cat_len(cat, drop) for cat in self.transformer.categories_]
        output_cols = int(sum(cat_lens))
        input_cols = len(self.transformer.categories_)
        feature_map = csr_matrix((input_cols, output_cols))
        last_num_cols = 0
        for i, cat_len in enumerate(cat_lens):
            feature_map[i, last_num_cols:(last_num_cols + cat_len)] = np.ones((cat_len))
            last_num_cols += cat_len

        self.set_feature_map(feature_map)

    def transform(self, x):
        """

        :param x: input data
        :type x: numpy.array
        :return: transformed data
        :rtype: numpy.array
        """
        ret = self.transformer.transform(x)
        if self._feature_map is None:
            self._build_feature_map()

        return ret


class ManytoManyMapper(FeatureMapper):
    def __init__(self, transformer):
        """Build a feature map for a many to many transformer.

        :param transformer: object that has a .transform method
        :type transformer: class that has a .transform method
        """
        super(ManytoManyMapper, self).__init__(transformer)
        self._transformer = transformer

    def _build_feature_map(self, in_feature_len=None, out_feature_len=None):
        """Generate a feature map so that weights of generated features are equally divided between parents and every
        parent feature map has every generated feature.

        :param in_feature_len: number of input features
        :type in_feature_len: int
        :param out_feature_len: number of output features
        :type out_feature_len: int
        """
        self.set_feature_map(np.ones((in_feature_len, out_feature_len)) * 1.0 / in_feature_len)

    def transform(self, x):
        """Get transformed data from input.

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: transformed data
        :rtype: numpy.array or scipy.sparse matrix
        """
        x_transformed = self._transformer.transform(x)
        if self._feature_map is None:
            self._build_feature_map(in_feature_len=x.shape[1], out_feature_len=x_transformed.shape[1])

        return x_transformed


class FuncTransformer:
    def __init__(self, func):
        """

        :param func: function that transforms the data
        :type func: function that takes in numpy.array/pandas.Dataframe and outputs numpy.array or scipy.sparse matrix
        """
        self._func = func

    def transform(self, x):
        """

        :param x: input data
        :type x: numpy.array or pandas.DataFrame
        :return: transformed data
        :rtype: numpy.array or scipy.sparse matrix
        """
        return self._func(x)


# dictionary containing currently identified preprocessors/transformers that result in one to many maps.
encoders_to_mappers_dict = {
    Binarizer: IdentityMapper,
    KernelCenterer: IdentityMapper,
    LabelEncoder: IdentityMapper,
    MaxAbsScaler: IdentityMapper,
    MinMaxScaler: IdentityMapper,
    Normalizer: IdentityMapper,
    QuantileTransformer: IdentityMapper,
    RobustScaler: IdentityMapper,
    StandardScaler: IdentityMapper,
    OneHotEncoder: OneHotEncoderMapper,
}

try:
    from sklearn.impute import MissingIndicator, SimpleImputer
    from sklearn.preprocessing import (KBinsDiscretizer, OrdinalEncoder,
                                       PowerTransformer)

    encoders_to_mappers_dict.update([
        (SimpleImputer, IdentityMapper),
        (KBinsDiscretizer, IdentityMapper),
        (MissingIndicator, IdentityMapper),
        (OrdinalEncoder, IdentityMapper),
        (PowerTransformer, IdentityMapper)
    ])
except ImportError:
    # sklearn version earlier than 0.20.0
    pass



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Common infrastructure, class hierarchy and utilities for model explanations."""

from .model_summary import ModelSummary

__all__ = ["ModelSummary"]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines metrics for validating model explanations."""

import numpy as np


def dcg(validate_order, ground_truth_order_relevance, top_values=10):
    """Compute the discounted cumulative gain (DCG).

    Compute the DCG as the sum of relevance scores penalized by the logarithmic position of the result.
    See https://en.wikipedia.org/wiki/Discounted_cumulative_gain for reference.

    :param validate_order: The order to validate.
    :type validate_order: list
    :param ground_truth_order_relevance: The ground truth relevancy of the documents to compare to.
    :type ground_truth_order_relevance: list
    :param top_values: Specifies the top values to compute the DCG for. The default is 10.
    :type top_values: int
    """
    # retrieve relevance score for each value in validation order
    relevance = np.vectorize(lambda x: ground_truth_order_relevance.get(x, 0))(validate_order[:top_values])
    gain = 2 ** relevance - 1
    discount = np.log2(np.arange(1, len(gain) + 1) + 1)
    sum_dcg = np.sum(gain / discount)
    return sum_dcg


def ndcg(validate_order, ground_truth_order, top_values=10):
    """Compute the normalized discounted cumulative gain (NDCG).

    Compute the NDCG as the ratio of the DCG for the validation order compared to the maximum DCG
    possible for the ground truth order.
    If the validation order is the same as the ground truth the NDCG will be the maximum of 1.0,
    and the least possible NDCG is 0.0.
    See https://en.wikipedia.org/wiki/Discounted_cumulative_gain for reference.

    :param validate_order: The order to validate for the documents. The values should be unique.
    :type validate_order: list
    :param ground_truth_order: The true order of the documents. The values should be unique.
    :type ground_truth_order: list
    :param top_values: Specifies the top values to compute the NDCG for. The default is 10.
    :type top_values: int
    """
    # Create map from true_order to "relevance" or reverse order index
    ground_truth_order_relevance = {}
    num_elems = len(ground_truth_order)
    for index, value in enumerate(ground_truth_order):
        # Compute the relevancy scores as a weighted reverse order in the given list.
        # Set the range of the relevance scores to be between 0 and 10
        # This is to prevent very large values when computing 2 ** relevance - 1
        ground_truth_order_relevance[value] = ((num_elems - index) / float(num_elems)) * 10.0
    # See https://en.wikipedia.org/wiki/Discounted_cumulative_gain for reference
    dcg_p = dcg(validate_order, ground_truth_order_relevance, top_values)
    idcg_p = dcg(ground_truth_order, ground_truth_order_relevance, top_values)
    ndcg = dcg_p / idcg_p
    return ndcg



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Reimports helpful model wrapper and utils for implicitly rewrapping the model to conform to explainer contracts."""

from ml_wrappers.model import (WrappedClassificationModel, WrappedPytorchModel,
                               WrappedRegressionModel, _wrap_model, wrap_model)

__all__ = ['WrappedClassificationModel', 'WrappedPytorchModel',
           'WrappedRegressionModel', '_wrap_model', 'wrap_model']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the DEPRECATED Explanation dashboard class."""

import warnings


class ExplanationDashboard:
    """DEPRECATED Explanation Dashboard class, please use the Explanation Dashboard from raiwidgets package instead.

    Since this class is deprecated it will no longer display the widget.
    Please install raiwidgets from pypi by running:
    pip install --upgrade raiwidgets
    The dashboard can be run with the same parameters in the new namespace:
    from raiwidgets import ExplanationDashboard

    :param explanation: An object that represents an explanation.
    :type explanation: ExplanationMixin
    :param model: An object that represents a model. It is assumed that for the classification case
        it has a method of predict_proba() returning the prediction probabilities for each
        class and for the regression case a method of predict() returning the prediction value.
    :type model: object
    :param dataset:  A matrix of feature vector examples (# examples x # features), the same samples
        used to build the explanation. Overwrites any existing dataset on the explanation object. Must have fewer than
        10000 rows and fewer than 1000 columns.
    :type dataset: numpy.array or list[][]
    :param datasetX: Alias of the dataset parameter. If dataset is passed, this will have no effect. Must have fewer
        than 10000 rows and fewer than 1000 columns.
    :type datasetX: numpy.array or list[][]
    :param true_y: The true labels for the provided dataset. Overwrites any existing dataset on the
        explanation object.
    :type true_y: numpy.array or list[]
    :param classes: The class names.
    :type classes: numpy.array or list[]
    :param features: Feature names.
    :type features: numpy.array or list[]
    :param port: The port to use on locally hosted service.
    :type port: int
    :param use_cdn: Deprecated. Whether to load latest dashboard script from cdn, fall back to local script if False.
        .. deprecated:: 0.15.2
           Deprecated since 0.15.2, cdn has been removed.  Setting parameter to True or False will trigger warning.
    :type use_cdn: bool
    :param public_ip: Optional. If running on a remote vm, the external public ip address of the VM.
    :type public_ip: str
    :param with_credentials: Optional. If running on a remote vm, sets up CORS policy both on client and server.
    :type with_credentials: bool
    """

    def __init__(self, explanation, model=None, *, dataset=None,
                 true_y=None, classes=None, features=None, port=None,
                 datasetX=None, trueY=None, locale=None, public_ip=None,
                 with_credentials=False, use_cdn=None):
        warnings.warn("ExplanationDashboard in interpret-community package is deprecated and removed."
                      "Please use the ExplanationDashboard from raiwidgets package instead.")



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the base explainer API to create explanations."""

from abc import ABC, abstractmethod

from .chained_identity import ChainedIdentity


class GlobalExplainer(ABC, ChainedIdentity):
    """The base class for explainers that create global explanations."""

    def __init__(self, *args, **kwargs):
        """Initialize the GlobalExplainer."""
        super(GlobalExplainer, self).__init__(*args, **kwargs)
        self._logger.debug('Initializing GlobalExplainer')

    @abstractmethod
    def explain_global(self, *args, **kwargs):
        """Abstract method to globally explain the given model.

        Note evaluation examples can be optional on derived classes since some explainers
        don't support it, for example MimicExplainer.

        :return: A model explanation object containing the global explanation.
        :rtype: GlobalExplanation
        """
        pass

    def __str__(self):
        """Get string representation of the explainer.

        :return: A string containing explainer name.
        :rtype: str
        """
        return "{}".format(self.__class__.__name__)


class LocalExplainer(ABC, ChainedIdentity):
    """The base class for explainers that create local explanations."""

    def __init__(self, *args, **kwargs):
        """Initialize the LocalExplainer."""
        super(LocalExplainer, self).__init__(*args, **kwargs)
        self._logger.debug('Initializing LocalExplainer')

    @abstractmethod
    def explain_local(self, evaluation_examples, **kwargs):
        """Abstract method to explain local instances.

        :param evaluation_examples: The evaluation examples.
        :type evaluation_examples: object
        :return: A model explanation object containing the local explanation.
        :rtype: LocalExplanation
        """
        pass

    def __str__(self):
        """Get string representation of the explainer.

        :return: A string containing explainer name.
        :rtype: str
        """
        return "{}".format(self.__class__.__name__)


class BaseExplainer(GlobalExplainer, LocalExplainer):
    """The base class for explainers that create global and local explanations."""

    def __init__(self, *args, **kwargs):
        """Initialize the BaseExplainer."""
        super(BaseExplainer, self).__init__(*args, **kwargs)
        self._logger.debug('Initializing BaseExplainer')



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Setup file for interpret-community package."""
import os
import shutil

from setuptools import find_packages, setup

with open('interpret_community/version.py') as f:
    code = compile(f.read(), f.name, 'exec')
    exec(code)

README_FILE = 'README.md'
LICENSE_FILE = 'LICENSE.txt'

# Note: used when generating the wheel but not on pip install of the package
if os.path.exists('../LICENSE'):
    shutil.copyfile('../LICENSE', LICENSE_FILE)


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS',
    'Operating System :: POSIX :: Linux'
]

DEPENDENCIES = [
    'numpy',
    'pandas',
    'scipy',
    'ml-wrappers==0.0.6',
    'scikit-learn',
    'packaging',
    'interpret-core[required]>=0.1.20, <=0.2.7',
    'shap>=0.20.0, <=0.40.0'
]

EXTRAS = {
    'sample': [
        'hdbscan'
    ],
    'deep': [
        'tensorflow',
        'pyyaml',
        'keras'
    ],
    'mimic': [
        'lightgbm'
    ],
    'lime': [
        'lime>=0.2.0.0'
    ]
}

with open(README_FILE, 'r', encoding='utf-8') as f:
    README = f.read()

setup(
    name=name,  # noqa: F821

    version=version,  # noqa: F821

    description='Microsoft Interpret Extensions SDK for Python',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Microsoft Corp',
    author_email='ilmat@microsoft.com',
    license='MIT License',
    url='https://github.com/interpretml/interpret-community',

    classifiers=CLASSIFIERS,

    packages=find_packages(exclude=["*.tests"]),

    install_requires=DEPENDENCIES,

    entry_points={
        "interpret_ext_blackbox": [
            "TabularExplainer = interpret_community:TabularExplainer",
            "KernelExplainer = interpret_community.shap:KernelExplainer",
            "MimicExplainer = interpret_community.mimic:MimicExplainer",
            "PFIExplainer = interpret_community.permutation:PFIExplainer",
            "LIMEExplainer = interpret_community.lime:LIMEExplainer"
        ],
        "interpret_ext_greybox": [
            "LinearExplainer = interpret_community.shap:LinearExplainer",
            "DeepExplainer = interpret_community.shap:DeepExplainer",
            "TreeExplainer = interpret_community.shap:TreeExplainer"
        ],
        "interpret_ext_glassbox": [
            "LGBMExplainableModel = interpret_community.mimic.models:LGBMExplainableModel",
            "LinearExplainableModel = interpret_community.mimic.models:LinearExplainableModel",
            "SGDExplainableModel = interpret_community.mimic.models:SGDExplainableModel",
            "DecisionTreeExplainableModel = interpret_community.mimic.models:DecisionTreeExplainableModel"
        ]
    },
    zip_safe=False,
    extras_require=EXTRAS
)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for mimic explainer and explainable surrogate models."""
from .mimic_explainer import MimicExplainer

__all__ = ["MimicExplainer"]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines utilities for handling kwargs on SHAP-based explainers."""

from interpret_community.common.constants import ExplainParams


def _get_explain_global_kwargs(sampling_policy, method, include_local, batch_size):
    """Get the kwargs for explain_global.

    :param sampling_policy: Optional policy for sampling the evaluation examples. See documentation on
        SamplingPolicy for more information.
    :type sampling_policy: interpret_community.common.SamplingPolicy
    :param method: The explanation method used, e.g., shap_kernel, mimic, etc.
    :type method: str
    :param include_local: Whether a local explanation should be generated or only global
    :type include_local: bool
    :param batch_size: If include_local is False, specifies the batch size for aggregating
        local explanations to global.
    :type batch_size: int
    :return: Args for explain_global.
    :rtype: dict
    """
    kwargs = {ExplainParams.METHOD: method,
              ExplainParams.SAMPLING_POLICY: sampling_policy,
              ExplainParams.INCLUDE_LOCAL: include_local,
              ExplainParams.BATCH_SIZE: batch_size}
    return kwargs



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for explainable surrogate models."""
from .explainable_model import BaseExplainableModel
from .lightgbm_model import LGBMExplainableModel
from .linear_model import LinearExplainableModel, SGDExplainableModel
from .tree_model import DecisionTreeExplainableModel

__all__ = ["BaseExplainableModel", "LGBMExplainableModel",
           "SGDExplainableModel", "LinearExplainableModel",
           "DecisionTreeExplainableModel"]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for Explanation Dashboard widget."""

from .explanation_dashboard import ExplanationDashboard

__all__ = ['ExplanationDashboard']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines utility functions for serialization of data."""

import datetime

import numpy as np


def _serialize_json_safe(o):
    """
    Convert a value into something that is safe to parse into JSON.

    :param o: Object to make JSON safe.
    :return: New object
    """
    if type(o) in {int, float, str, type(None)}:
        if isinstance(o, float):
            if np.isinf(o) or np.isnan(o):
                return 0
        return o
    elif isinstance(o, datetime.datetime):
        return o.__str__()
    elif isinstance(o, dict):
        return {k: _serialize_json_safe(v) for k, v in o.items()}
    elif isinstance(o, list):
        return [_serialize_json_safe(v) for v in o]
    elif isinstance(o, tuple):
        return tuple(_serialize_json_safe(v) for v in o)
    elif isinstance(o, np.ndarray):
        return _serialize_json_safe(o.tolist())
    else:
        # Attempt to convert Numpy type
        try:
            return o.item()
        except Exception:
            return o



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a decorator for tabular data which wraps pandas dataframes, scipy and numpy arrays in a DatasetWrapper."""

from functools import wraps

from .dataset_wrapper import DatasetWrapper


def tabular_decorator(explain_func):
    """Decorate an explanation function to wrap evaluation examples in a DatasetWrapper.

    :param explain_func: An explanation function where the first argument is a dataset.
    :type explain_func: explanation function
    """
    @wraps(explain_func)
    def explain_func_wrapper(self, evaluation_examples, *args, **kwargs):
        if not isinstance(evaluation_examples, DatasetWrapper):
            self._logger.debug('Eval examples not wrapped, wrapping')
            evaluation_examples = DatasetWrapper(evaluation_examples)
        return explain_func(self, evaluation_examples, *args, **kwargs)
    return explain_func_wrapper


def wrap_dataset(dataset):
    if not isinstance(dataset, DatasetWrapper):
        dataset = DatasetWrapper(dataset)
    return dataset


def init_tabular_decorator(init_func):
    """Decorate a constructor to wrap initialization examples in a DatasetWrapper.

    Provided for convenience for tabular data explainers.

    :param init_func: Initialization constructor where the second argument is a dataset.
    :type init_func: Initialization constructor.
    """
    @wraps(init_func)
    def init_wrapper(self, model, initialization_examples, *args, **kwargs):
        return init_func(self, model, wrap_dataset(initialization_examples), *args, **kwargs)
    return init_wrapper



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Contains functions useful for data mapper."""

import pandas


def _get_list_from_slice(s):
    """Get a list from a slice object.

    :param s: slice object representing list of columns
    :type s: slice
    :return: list of elements that were in the slice object
    :rtype: list[int]
    """
    args = [s.start, s.stop]
    if s.step is not None:
        args.append(s.step)
    return list(range(*args))


def extract_column(x, transformer_config):
    """Extract column from DataFrame/numpy array x and reshape if column_name is list.

    :param x: input raw data
    :type x: numpy.array or pandas.DataFrame
    :param transformer_config: TransformerConfig used to reshape input data
    :type transformer_config: TransformerConfig
    :return: column or columns from input data
    :rtype: numpy.array
    """
    columns = transformer_config.columns

    if isinstance(x, pandas.DataFrame):
        x_column = x[columns].values
    else:
        x_column = x[:, columns]
    if len(x.shape) == 1 and not transformer_config.one_dimensional:
        x_column = x_column.reshape(-1, 1)

    return x_column


def get_transformer_config_tuples_from_transformations_list(transformations):
    """Get transformer and transformer config tuples from list of transformations in sklearn-pandas format.

    :param transformations: list of transformations to be applied to input data in sklearn-pandas format
    :type transformations: list[tuple]
    :return: list of two element tuples. First element is a transformer object and second element is a
    TransformerConfig object
    :rtype: list[tuple]
    """
    tuples = []
    for columns, transformer in transformations:
        one_dimensional = False
        if isinstance(columns, str) or isinstance(columns, int):
            one_dimensional = True
            columns = [columns]
        tuples.append((transformer, TransformerConfig(columns, one_dimensional)))

    return tuples


def get_transformer_config_tuples_from_column_transformer(column_transformer):
    """Get tuples of transformer, transformer config from sklearn.compose.ColumnTransformer

    :param column_transformer: ColumnTransformer that transforms the input data
    :type column_transformer: sklearn.compose.ColumnTransformer
    :return: list of tuples of the form of (transformer, TransformerConfig)
    :rtype: list[tuple]
    """
    tuples = []
    for _, transformer, columns in column_transformer.transformers_:
        if transformer == "drop":
            continue
        elif transformer == "passthrough":
            transformer = None

        if isinstance(columns, slice):
            columns = _get_list_from_slice(columns)

        one_dimensional = False
        if isinstance(columns, str) or isinstance(columns, int):
            one_dimensional = True
            columns = [columns]

        tuples.append((transformer, TransformerConfig(columns, one_dimensional)))

    return tuples


class TransformerConfig:
    """Class that stores information about the columns on which a transformation operates."""

    def __init__(self, columns=None, one_dimensional=False):
        """Initialize the TransformerConfig object.

        :param one_dimensional: boolean flag to indicate whether the data is one dimensional.
        :type one_dimensional: bool
        :param columns: list of input columns
        :type columns: [str]
        """
        self._one_dimensional = one_dimensional
        self._columns = columns

    @property
    def one_dimensional(self):
        """Bool that indicates whether the given transformer expects one dimensional input."""
        return self._one_dimensional

    @property
    def columns(self):
        """List of columns which the transformer takes as input."""
        return self._columns



import os
from tempfile import TemporaryDirectory

import interpret_community
import yaml

from ..explanation.serialization import (_get_explanation_metadata,
                                         load_explanation, save_explanation)


def _load_pyfunc(path):
    """Load the explanation from the given path.

    :param path: The path from which to load the explanation.
    :type path: str
    """

    load_explanation(path)


def save_model(path, loader_module=None, data_path=None, conda_env=None, mlflow_model=None, **kwargs):
    """Save the explanation locally using the MLflow model format.

    This function is necessary for log_explanation to work properly.

    :param path: The destination path for the saved explanation.
    :type path: str
    :param loader_module: The package that will be used to reload a serialized explanation. In this case,
        always interpret_community.mlflow.
    :type loader_module: str
    :param data_path: The path to the serialized explanation files.
    :type data_path: str
    :param conda_env: The path to a YAML file with basic Python environment information.
    :type conda_env: str
    :param mlflow_model: In our case, always None.
    :type mlflow_model: None
    :return: The MLflow model representation of the explanation.
    :rtype: mlflow.models.Model
    """

    try:
        import mlflow
        from mlflow.models import Model
        from mlflow.pyfunc.model import get_default_conda_env
        from mlflow.utils.file_utils import _copy_file_or_tree
    except ImportError as e:
        raise Exception("Could not log_explanation to mlflow. Missing mlflow dependency, "
                        "pip install mlflow to resolve the error: {}.".format(e))

    if os.path.exists(path):
        raise Exception(
            message="Path '{}' already exists".format(path))
    os.makedirs(path)

    if mlflow_model is None:
        mlflow_model = Model()

    data = None
    if data_path is not None:
        model_file = _copy_file_or_tree(src=data_path, dst=path, dst_dir="data")
        data = model_file

    conda_env_subpath = "mlflow_env.yml"
    if conda_env is None:
        conda_env = get_default_conda_env()
    elif not isinstance(conda_env, dict):
        with open(conda_env, "r") as f:
            conda_env = yaml.safe_load(f)
    with open(os.path.join(path, conda_env_subpath), "w") as f:
        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)

    mlflow.pyfunc.add_to_model(
        mlflow_model, loader_module=loader_module, data=data, env=conda_env_subpath, **kwargs)
    mlflow_model.save(os.path.join(path, 'MLmodel'))
    return mlflow_model


def log_explanation(name, explanation):
    """Log the explanation to MLflow using MLflow model logging.

    :param name: The name of the explanation. Will be used as a directory name.
    :type name: str
    :param explanation: The explanation object to log.
    :type explanation: Explanation
    """

    try:
        from mlflow.models import Model
    except ImportError as e:
        raise Exception("Could not log_explanation to mlflow. Missing mlflow dependency, "
                        "pip install mlflow to resolve the error: {}.".format(e))
    import cloudpickle as pickle

    with TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, 'explanation')
        save_explanation(explanation, path, exist_ok=True)

        conda_env = {
            "name": "mlflow-env",
            "channels": ["defaults"],
            "dependencies": [
                "pip",
                {
                    "pip": [
                        "interpret-community=={}".format(interpret_community.__version__),
                        "cloudpickle=={}".format(pickle.__version__)]
                }
            ]
        }
        conda_path = os.path.join(tempdir, "conda.yaml")
        with open(conda_path, "w") as stream:
            yaml.dump(conda_env, stream)
        kwargs = {'interpret_community_metadata': _get_explanation_metadata(explanation)}
        Model.log(name,
                  flavor=interpret_community.mlflow,
                  loader_module='interpret_community.mlflow',
                  data_path=path,
                  conda_env=conda_path,
                  **kwargs)


def get_explanation(run_id, name):
    """Download and deserialize an explanation that has been logged to MLflow.

    :param run_id: The ID of the run the explanation was logged to.
    :type run_id: str
    :param name: The name given to the explanation when it was logged.
    :type name: str
    :return: The rehydrated explanation.
    :rtype: Explanation
    """

    try:
        import mlflow
    except ImportError as e:
        raise Exception("Could not get_explanation from mlflow. Missing mlflow dependency, "
                        "pip install mlflow to resolve the error: {}.".format(e))
    DOWNLOAD_DIR = 'exp_downloads'
    client = mlflow.tracking.MlflowClient()
    os.makedirs(DOWNLOAD_DIR, exist_ok=True)
    client.download_artifacts(run_id, name, dst_path=DOWNLOAD_DIR)
    full_path = os.path.join(DOWNLOAD_DIR, name, 'data', 'explanation')
    return load_explanation(full_path)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the save and load explanation methods and helpers for serialization."""

import json
import os

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, issparse

from ..common.constants import ExplainParams, ExplainType
from ..dataset.dataset_wrapper import DatasetWrapper
from ..explanation import explanation as patched_explanation
from .explanation import (ClassesMixin, FeatureImportanceExplanation,
                          GlobalExplanation, LocalExplanation,
                          _create_global_explanation,
                          _create_local_explanation)

EXPLANATION_METADATA = 'explanation_metadata.json'
SERIALIZATION_VERSION = 'serialization_version'


def save_explanation(explanation, path, exist_ok=False):
    """Serialize the explanation.

    :param explanation: The Explanation to be serialized.
    :type explanation: Explanation
    :param path: The path to the directory in which the explanation will be saved. By default, must be a new directory
        to avoid overwriting any previous explanations. Set exist_ok to True to overrule this behavior.
    :type path: str
    :param exist_ok: If False (default), the path provided by the user must not already exist and will be created by
        this function. If True, a prexisting path may be passed. Any preexisting files whose names match those of the
        files that make up the explanation will be overwritten.
    :type exist_ok: bool
    """
    if os.path.exists(path) and not exist_ok:
        raise Exception('The directory specified by path already exists. '
                        'Please pass in a new directory or set exists_ok=True.')
    os.makedirs(path, exist_ok=True)

    # TODO replace with set of params from below
    uploadable_properties = [
        ExplainParams.FEATURES,
        ExplainParams.LOCAL_IMPORTANCE_VALUES,
        ExplainParams.EXPECTED_VALUES,
        ExplainParams.CLASSES,
        ExplainParams.GLOBAL_IMPORTANCE_NAMES,
        ExplainParams.GLOBAL_IMPORTANCE_RANK,
        ExplainParams.GLOBAL_IMPORTANCE_VALUES,
        ExplainParams.PER_CLASS_NAMES,
        ExplainParams.PER_CLASS_RANK,
        ExplainParams.PER_CLASS_VALUES,
        ExplainParams.INIT_DATA,
        ExplainParams.EVAL_DATA,
        ExplainParams.EVAL_Y_PRED,
        ExplainParams.EVAL_Y_PRED_PROBA
    ]
    # TODO will need to add viz data on top of this
    for prop in uploadable_properties:
        if hasattr(explanation, prop) and getattr(explanation, prop) is not None:
            value = getattr(explanation, prop)
            if isinstance(value, pd.DataFrame):
                value = value.values.tolist()
                metadata = 'DataFrame'
            elif isinstance(value, DatasetWrapper):
                value = value.original_dataset
                if issparse(value):
                    value = _convert_sparse_to_list(value)
                    metadata = 'DatasetWrapperSparse'
                else:
                    value = value.tolist()
                    metadata = 'DatasetWrapper'
            elif isinstance(value, np.ndarray):
                value = value.tolist()
                metadata = 'ndarray'
            elif isinstance(value, list):
                metadata = 'list'
            elif issparse(value):
                if prop != ExplainParams.LOCAL_IMPORTANCE_VALUES:
                    value = _convert_sparse_to_list(value)
                metadata = 'sparse'
            else:
                metadata = 'other'
            if prop == ExplainParams.LOCAL_IMPORTANCE_VALUES and explanation.is_local_sparse:
                value = _convert_sparse_data(value, explanation)
                metadata = 'sparse'
            data_dict = {
                'metadata': metadata,
                'data': value
            }
            filename = os.path.join(path, prop + '.json')
            with open(filename, 'w') as f:
                json.dump(data_dict, f)
    # create metadata file
    prop_dict = _get_explanation_metadata(explanation)

    filename = os.path.join(path, EXPLANATION_METADATA)
    with open(filename, 'w') as f:
        json.dump(prop_dict, f)


def load_explanation(path):
    """Deserialize the explanation.

    :param path: The path to the directory in which the explanation will be saved. By default, must be a new directory
        to avoid overwriting any previous explanations. Set exist_ok to True to overrule this behavior.
    :type path: str
    :return: The deserialized explanation.
    :rtype: Explanation
    """
    shared_params = [
        ExplainParams.EXPECTED_VALUES,
        ExplainParams.FEATURES,
        ExplainParams.CLASSES,
        ExplainParams.EVAL_DATA,
        ExplainParams.INIT_DATA,
        ExplainParams.EVAL_Y_PRED,
        ExplainParams.EVAL_Y_PRED_PROBA
    ]
    global_params = [
        ExplainParams.GLOBAL_IMPORTANCE_RANK,
        ExplainParams.GLOBAL_IMPORTANCE_VALUES,
        ExplainParams.PER_CLASS_RANK,
        ExplainParams.PER_CLASS_VALUES
    ]
    local_params = [
        ExplainParams.LOCAL_IMPORTANCE_VALUES
    ]
    with open(os.path.join(path, EXPLANATION_METADATA), 'r') as f:
        metadata = json.load(f)
        is_global = metadata[ExplainType.GLOBAL]
        is_local = metadata[ExplainType.LOCAL]

    if is_local and is_global:
        local_kwargs = _get_kwargs(path, shared_params + local_params)
        local_explanation = _create_local_explanation(**local_kwargs)
        global_kwargs = _get_kwargs(path, shared_params + global_params, local_explanation=local_explanation)
        return _create_global_explanation(**global_kwargs)
    elif is_local:
        local_kwargs = _get_kwargs(path, shared_params + local_params)
        return _create_local_explanation(**local_kwargs)
    elif is_global:
        global_kwargs = _get_kwargs(path, shared_params + global_params)
        return _create_global_explanation(**global_kwargs)


# Monkey patch the save and load explanation methods in explanation.py for backwards compat
patched_explanation.save_explanation = save_explanation
patched_explanation.load_explanation = load_explanation


def _get_explanation_metadata(explanation):
    classification = ClassesMixin._does_quack(explanation)
    is_raw = False if not FeatureImportanceExplanation._does_quack(explanation) else explanation.is_raw
    is_eng = False if not FeatureImportanceExplanation._does_quack(explanation) else explanation.is_engineered
    num_features = 0 if not FeatureImportanceExplanation._does_quack(explanation) else explanation.num_features
    prop_dict = {
        ExplainParams.EXPLANATION_ID: explanation.id,
        ExplainType.MODEL: ExplainType.CLASSIFICATION if classification else ExplainType.REGRESSION,
        ExplainType.DATA: ExplainType.TABULAR,
        ExplainType.MODEL_TASK: explanation.model_task,
        ExplainType.METHOD: explanation.method,
        ExplainType.MODEL_CLASS: explanation.model_type,
        ExplainType.IS_RAW: is_raw,
        ExplainType.IS_ENG: is_eng,
        ExplainType.GLOBAL: GlobalExplanation._does_quack(explanation),
        ExplainType.LOCAL: LocalExplanation._does_quack(explanation),
        ExplainParams.NUM_CLASSES: 1 if not ClassesMixin._does_quack(explanation) else explanation.num_classes,
        ExplainParams.NUM_EXAMPLES: 0 if not LocalExplanation._does_quack(explanation) else explanation.num_examples,
        ExplainParams.NUM_FEATURES: num_features,
        ExplainParams.IS_LOCAL_SPARSE: LocalExplanation._does_quack(explanation) and explanation.is_local_sparse,
        SERIALIZATION_VERSION: 1,
    }
    return prop_dict


def _get_value_from_file(file_var):
    json_input = json.load(file_var)
    meta = json_input['metadata']
    data = json_input['data']
    if meta == 'DataFrame':
        return pd.DataFrame(data)
    elif meta == 'DatasetWrapper':
        if isinstance(data, list):
            data = np.array(data)
        return DatasetWrapper(data)
    elif meta == 'DatasetWrapperSparse':
        data = _convert_artifact_to_sparse_local(data)
        return DatasetWrapper(data)
    elif meta == 'ndarray':
        return np.array(data)
    elif meta == 'list':
        return data
    elif meta == 'sparse':
        return _convert_artifact_to_sparse_local(data)
    else:
        raise Exception('Unrecognized data type in deserialization: {}'.format(meta))


def _get_kwargs(path, params, local_explanation=None):
    with open(os.path.join(path, EXPLANATION_METADATA), 'r') as f:
        metadata = json.load(f)
    numpy_params = {ExplainParams.EVAL_DATA, ExplainParams.INIT_DATA}
    is_local_sparse = metadata[ExplainParams.IS_LOCAL_SPARSE]
    if not is_local_sparse:
        numpy_params.add(ExplainParams.LOCAL_IMPORTANCE_VALUES)

    kwargs = {}
    for param in params:
        if os.path.exists(os.path.join(path, param + '.json')):
            with open(os.path.join(path, param + '.json'), 'r') as f:
                kwargs[param] = _get_value_from_file(f)
                if param in numpy_params:
                    kwargs[param] = np.array(kwargs[param])

    param_list = [
        ExplainParams.METHOD,
        ExplainParams.MODEL_TASK,
        ExplainParams.NUM_FEATURES,
        ExplainParams.IS_RAW,
        ExplainParams.IS_ENG,
        ExplainParams.EXPLANATION_ID
    ]
    for param in param_list:
        kwargs[param] = metadata[param]
    kwargs[ExplainParams.MODEL_TYPE] = metadata[ExplainType.MODEL_CLASS]
    classification = metadata[ExplainType.MODEL] == ExplainType.CLASSIFICATION
    kwargs[ExplainParams.CLASSIFICATION] = classification
    if classification:
        kwargs[ExplainParams.NUM_CLASSES] = metadata[ExplainParams.NUM_CLASSES]
    if local_explanation is not None:
        kwargs[ExplainParams.LOCAL_EXPLANATION] = local_explanation
    return kwargs


def _convert_sparse_data(value, explanation):
    """Converts the given sparse local importance values to list format.

    :param value: The local importance values to convert.
    :type value: list[scipy.sparse.csr_matrix] or scipy.sparse.csr_matrix
    :return: The local importance values as a list representation.
    :rtype: list[list[int | float]]
    """
    if issparse(value):
        return _convert_sparse_to_list(value)
    else:
        sparse_data = []
        for sparse_local_values in value:
            sparse_data.append(_convert_sparse_to_list(sparse_local_values))
        return sparse_data


def _convert_sparse_to_list(sparse_local_values):
    """Converts a sparse explanation to a list.

    :param sparse_local_values: The sparse matrix to be converted to a list representation that includes
        the data, indices, indptr and shape.
    :type sparse_local_values: scipy.sparse.csr_matrix
    :return: The local importance values as a list representation.
    :rtype: list[list[int | float]]
    """
    sparse_data = []
    sparse_data.append(sparse_local_values.data.tolist())
    sparse_data.append(sparse_local_values.indices.tolist())
    sparse_data.append(sparse_local_values.indptr.tolist())
    sparse_data.append(list(sparse_local_values.shape))
    return sparse_data


def _convert_artifact_to_sparse_local(local_vals_sparse):
    """Converts a saved representation of an explanation to sparse local importance values.

    :param local_vals_sparse: The list representation that includes
        the data, indices, indptr and shape to be converted to a sparse matrix.
        In multiclass case this includes the per class local importance values
        as well, so it will be a 3 dimensional list of values.
    :type local_vals_sparse: list[list[int | float]] or list[list[list[int | float]]]
    :return: The local importance values as a sparse matrix.
    :rtype: list[scipy.sparse.csr_matrix] or scipy.sparse.csr_matrix
    """
    if isinstance(local_vals_sparse[0][0], list):
        local_importance_vals = []
        for class_importance_values in local_vals_sparse:
            local_importance_vals.append(_convert_list_to_sparse(class_importance_values))
    else:
        local_importance_vals = _convert_list_to_sparse(local_vals_sparse)
    return local_importance_vals


def _convert_list_to_sparse(sparse_local_values_list):
    """Converts a list representation of an explanation to a sparse matrix.

    :param sparse_local_values_list: The list representation that includes
        the data, indices, indptr and shape to be converted to a sparse matrix.
    :type sparse_local_values_list: list[list[int | float]]
    :return: The local importance values as a sparse matrix.
    :rtype: scipy.sparse.csr_matrix
    """
    data = sparse_local_values_list[0]
    indices = sparse_local_values_list[1]
    indptr = sparse_local_values_list[2]
    shape = tuple(sparse_local_values_list[3])
    return csr_matrix((data, indices, indptr), shape)



# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
The code is based on the similar utility function from SHAP:
https://github.com/slundberg/shap/blob/9411b68e8057a6c6f3621765b89b24d82bee13d4/shap/utils/_legacy.py
This version makes use of cuml kmeans instead of sklearn for speed.
"""

import numpy as np

try:
    import cuml
    from cuml import KMeans
    from cuml.preprocessing import SimpleImputer
    rapids_installed = True
except BaseException:
    rapids_installed = False
from scipy.sparse import issparse


def kmeans(X, k, round_values=True):
    """ Summarize a dataset with k mean samples weighted by the number of data points they
    each represent.
    Parameters
    ----------
    X : numpy.array or pandas.DataFrame or any scipy.sparse matrix
        Matrix of data samples to summarize (# samples x # features)
    k : int
        Number of means to use for approximation.
    round_values : bool
        For all i, round the ith dimension of each mean sample to match the nearest value
        from X[:,i]. This ensures discrete features always get a valid value.
    Returns
    -------
    DenseData object.
    """

    if not rapids_installed:
        raise RuntimeError(
            "cuML is required to use GPU explainers. Check https://rapids.ai/start.html \
            for more information on how to install it.")
    if cuml.__version__ >= '21.08':
        from cuml.explainer.sampling import kmeans_sampling
        summary, group_names, labels = kmeans_sampling(X, k, round_values, detailed=True)

        return DenseData(summary,
                         group_names,
                         None,
                         1.0 * np.bincount(labels))
    # For backward compatibility
    group_names = [str(i) for i in range(X.shape[1])]
    if str(type(X)).endswith("'pandas.core.frame.DataFrame'>"):
        group_names = X.columns
        X = X.values

    # in case there are any missing values in data impute them
    imp = SimpleImputer(missing_values=np.nan, strategy='mean')
    X = imp.fit_transform(X)

    kmeans = KMeans(n_clusters=k, random_state=0).fit(X)

    if round_values:
        for i in range(k):
            for j in range(X.shape[1]):
                xj = X[:, j].toarray().flatten() if issparse(
                    X) else X[:, j]  # sparse support courtesy of @PrimozGodec
                ind = np.argmin(np.abs(xj - kmeans.cluster_centers_[i, j]))
                kmeans.cluster_centers_[i, j] = X[ind, j]
    return DenseData(
        kmeans.cluster_centers_,
        group_names,
        None,
        1.0 *
        np.bincount(
            kmeans.labels_))


class Data:
    def __init__(self):
        pass


class DenseData(Data):
    def __init__(self, data, group_names, *args):
        self.groups = args[0] if len(args) > 0 and args[0] is not None else [
            np.array([i]) for i in range(len(group_names))]

        length = sum(len(g) for g in self.groups)
        num_samples = data.shape[0]
        t = False
        if length != data.shape[1]:
            t = True
            num_samples = data.shape[1]

        valid = (
            not t and length == data.shape[1]) or (
            t and length == data.shape[0])
        assert valid, "# of names must match data matrix!"

        self.weights = args[1] if len(args) > 1 else np.ones(num_samples)
        self.weights /= np.sum(self.weights)
        wl = len(self.weights)
        valid = (not t and wl == data.shape[0]) or (t and wl == data.shape[1])
        assert valid, "# weights must match data matrix!"

        self.transposed = t
        self.group_names = group_names
        self.data = data
        self.groups_size = len(self.groups)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for interaction with MLflow."""

from .mlflow import get_explanation, log_explanation, save_model

__all__ = ['get_explanation', 'log_explanation', 'save_model']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the Mimic Explainer for computing explanations on black box models or functions.

The mimic explainer trains an explainable model to reproduce the output of the given black box model.
The explainable model is called a surrogate model and the black box model is called a teacher model.
Once trained to reproduce the output of the teacher model, the surrogate model's explanation can
be used to explain the teacher model.
"""

import json
import logging
import warnings

import numpy as np
from scipy.sparse import issparse
from sklearn.metrics import accuracy_score, r2_score

from .._internal.raw_explain.raw_explain_utils import (
    get_datamapper_and_transformed_data, transform_with_datamapper)
from ..common.blackbox_explainer import BlackBoxExplainer
from ..common.constants import (Defaults, ExplainableModelType, ExplainParams,
                                ExplainType, Extension, LightGBMParams,
                                MimicSerializationConstants, ModelTask,
                                ResetIndex, ShapValuesOutput)
from ..common.exception import ScenarioNotSupportedException
from ..common.explanation_utils import _order_imp
from ..common.model_wrapper import _wrap_model
from ..dataset.dataset_wrapper import DatasetWrapper
from ..dataset.decorator import init_tabular_decorator, tabular_decorator
from ..explanation.explanation import (
    _aggregate_global_from_local_explanation,
    _aggregate_streamed_local_explanations, _create_global_explanation,
    _create_local_explanation, _create_raw_feats_global_explanation,
    _create_raw_feats_local_explanation,
    _get_raw_explainer_create_explanation_kwargs)
from .model_distill import _inverse_soft_logit, _model_distill
from .models import LGBMExplainableModel

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', 'Starting from version 2.2.1', UserWarning)


class MimicExplainer(BlackBoxExplainer):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.BLACKBOX

    """The Mimic Explainer for explaining black box models or functions.

    :param model: The black box model or function (if is_function is True) to be explained. Also known
        as the teacher model.
        A model that implements sklearn.predict or sklearn.predict_proba or function that accepts a 2d ndarray.
    :type model: object
    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param explainable_model: The uninitialized surrogate model used to explain the black box model.
        Also known as the student model.
    :type explainable_model: interpret_community.mimic.models.BaseExplainableModel
    :param explainable_model_args: An optional map of arguments to pass to the explainable model
        for initialization.
    :type explainable_model_args: dict
    :param is_function: Default is False. Set to True if passing function instead of model.
    :type is_function: bool
    :param augment_data: If True, oversamples the initialization examples to improve surrogate
        model accuracy to fit teacher model. Useful for high-dimensional data where
        the number of rows is less than the number of columns.
    :type augment_data: bool
    :param max_num_of_augmentations: Maximum number of times we can increase the input data size.
    :type max_num_of_augmentations: int
    :param explain_subset: List of feature indices. If specified, only selects a subset of the
        features in the evaluation dataset for explanation. Note for mimic explainer this will
        not affect the execution time of getting the global explanation. This argument is not supported when
        transformations are set.
    :type explain_subset: list[int]
    :param features: A list of feature names.
    :type features: list[str]
    :param classes: Class names as a list of strings. The order of the class names should match
        that of the model output. Only required if explaining classifier.
    :type classes: list[str]
    :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
        transformer. When transformations are provided, explanations are of the features before the transformation.
        The format for a list of transformations is same as the one here:
        https://github.com/scikit-learn-contrib/sklearn-pandas.

        If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
        are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
        package, then this parameter cannot take a list of more than one column as input for the transformation.
        You can use the following sklearn.preprocessing  transformations with a list of columns since these are
        already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
        MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
        RobustScaler, StandardScaler.

        Examples for transformations that work::

            [
                (["col1", "col2"], sklearn_one_hot_encoder),
                (["col3"], None) #col3 passes as is
            ]
            [
                (["col1"], my_own_transformer),
                (["col2"], my_own_transformer),
            ]

        An example of a transformation that would raise an error since it cannot be interpreted as one to many::

            [
                (["col1", "col2"], my_own_transformer)
            ]

        The last example would not work since the interpret-community package can't determine whether
        my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
    :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
    :param shap_values_output: The shap values output from the explainer.  Only applies to
        tree-based models that are in terms of raw feature values instead of probabilities.
        Can be default, probability or teacher_probability. If probability or teacher_probability
        are specified, we approximate the feature importance values as probabilities instead
        of using the default values. If teacher probability is specified, we use the probabilities
        from the teacher model as opposed to the surrogate model.
    :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
    :param categorical_features: Categorical feature names or indexes.
        If names are passed, they will be converted into indexes first.
        Note if pandas indexes are categorical, you can either pass the name of the index or the index
        as if the pandas index was inserted at the end of the input dataframe.
    :type categorical_features: Union[list[str], list[int]]
    :param allow_all_transformations: Allow many to many and many to one transformations
    :type allow_all_transformations: bool
    :param model_task: Optional parameter to specify whether the model is a classification or regression model.
        In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
        has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
        outputs a 1 dimensional array.
    :type model_task: str
    :param reset_index: Uses the pandas DataFrame index column as part of the features when training
        the surrogate model.
    :type reset_index: str
    """

    @init_tabular_decorator
    def __init__(self, model, initialization_examples, explainable_model, explainable_model_args=None,
                 is_function=False, augment_data=True, max_num_of_augmentations=10, explain_subset=None,
                 features=None, classes=None, transformations=None, allow_all_transformations=False,
                 shap_values_output=ShapValuesOutput.DEFAULT, categorical_features=None,
                 model_task=ModelTask.Unknown, reset_index=ResetIndex.Ignore, **kwargs):
        """Initialize the MimicExplainer.

        :param model: The black box model or function (if is_function is True) to be explained.  Also known
            as the teacher model.
            A model that implements sklearn.predict or sklearn.predict_proba or function that accepts a 2d
            ndarray.
        :type model: object
        :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param explainable_model: The uninitialized surrogate model used to explain the black box model.
            Also known as the student model.
        :type explainable_model: BaseExplainableModel
        :param explainable_model_args: An optional map of arguments to pass to the explainable model
            for initialization.
        :type explainable_model_args: dict
        :param is_function: Default is False. Set to True if passing function instead of model.
        :type is_function: bool
        :param augment_data: If True, oversamples the initialization examples to improve surrogate
            model accuracy to fit teacher model.  Useful for high-dimensional data where
            the number of rows is less than the number of columns.
        :type augment_data: bool
        :param max_num_of_augmentations: Maximum number of times we can increase the input data size.
        :type max_num_of_augmentations: int
        :param explain_subset: List of feature indices. If specified, only selects a subset of the
            features in the evaluation dataset for explanation. Note for mimic explainer this will
            not affect the execution time of getting the global explanation. This argument is not supported when
            transformations are set.
        :type explain_subset: list[int]
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
            transformer. When transformations are provided, explanations are of the features before the transformation.
            The format for a list of transformations is same as the one here:
            https://github.com/scikit-learn-contrib/sklearn-pandas.

            If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
            are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
            package, then this parameter cannot take a list of more than one column as input for the transformation.
            You can use the following sklearn.preprocessing  transformations with a list of columns since these are
            already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
            MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
            RobustScaler, StandardScaler.

            Examples for transformations that work::

                [
                    (["col1", "col2"], sklearn_one_hot_encoder),
                    (["col3"], None) #col3 passes as is
                ]
                [
                    (["col1"], my_own_transformer),
                    (["col2"], my_own_transformer),
                ]

            An example of a transformation that would raise an error since it cannot be interpreted as one to many::

                [
                    (["col1", "col2"], my_own_transformer)
                ]

            The last example would not work since the interpret-community package can't determine whether
            my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
        :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
        :param shap_values_output: The shap values output from the explainer.  Only applies to
            tree-based models that are in terms of raw feature values instead of probabilities.
            Can be default, probability or teacher_probability. If probability or teacher_probability
            are specified, we approximate the feature importance values as probabilities instead
            of using the default values. If teacher probability is specified, we use the probabilities
            from the teacher model as opposed to the surrogate model.
        :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
        :param categorical_features: Categorical feature names or indexes.
            If names are passed, they will be converted into indexes first.
            Note if pandas indexes are categorical, you can either pass the name of the index or the index
            as if the pandas index was inserted at the end of the input dataframe.
        :type categorical_features: Union[list[str], list[int]]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :param model_task: Optional parameter to specify whether the model is a classification or regression model.
            In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
            has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
            outputs a 1 dimensional array.
        :type model_task: str
        :param reset_index: Can be ignore, reset or reset_teacher.  By default we ignore the index column, but the
            user can override to reset it and make it a feature column that is then featurized to numeric. Or,
            when using reset_teacher, the user can reset it and ignore it during featurization but set it as
            the index when calling predict on the original model.  Only use reset_teacher if the index is already
            featurized as part of the data.
        :type reset_index: str
        """
        if transformations is not None and explain_subset is not None:
            raise ValueError("explain_subset not supported with transformations")
        self.reset_index = reset_index
        self._datamapper = None
        if transformations is not None:
            self._datamapper, new_initialization_examples = get_datamapper_and_transformed_data(
                examples=initialization_examples, transformations=transformations,
                allow_all_transformations=allow_all_transformations)
            # memory optimization to clear all internal references so they can be gc'ed as soon as possible
            initialization_examples._clear()
            initialization_examples = new_initialization_examples
        if reset_index != ResetIndex.Ignore:
            initialization_examples.reset_index()
        wrapped_model, eval_ml_domain = _wrap_model(model, initialization_examples, model_task, is_function)
        super(MimicExplainer, self).__init__(wrapped_model, is_function=is_function,
                                             model_task=eval_ml_domain, **kwargs)
        if explainable_model_args is None:
            explainable_model_args = {}
        if categorical_features is None:
            categorical_features = []
        self._logger.debug('Initializing MimicExplainer')

        # Get the feature names from the initialization examples
        self._init_features = initialization_examples.get_features(features=features)
        self.features = features
        # augment the data if necessary
        if augment_data:
            initialization_examples.augment_data(max_num_of_augmentations=max_num_of_augmentations)
            # get the original data with types and index column if reset_index != ResetIndex.Ignore
            original_training_data = initialization_examples.typed_dataset
        else:
            # get the original data with types and index column if reset_index != ResetIndex.Ignore
            # note: this is small memory optimization over above line if augment data = False
            original_training_data = initialization_examples.original_dataset_with_type

        # if index column should not be set on surrogate model, remove it
        if reset_index == ResetIndex.ResetTeacher:
            initialization_examples.set_index()

        # If categorical_features is a list of string column names instead of indexes, make sure to convert to indexes
        if not all(isinstance(categorical_feature, int) for categorical_feature in categorical_features):
            categorical_features = initialization_examples.get_column_indexes(self._init_features,
                                                                              categorical_features)

        # Featurize any timestamp columns
        # TODO: more sophisticated featurization
        self._timestamp_featurizer = initialization_examples.timestamp_featurizer()

        # If model is a linear model or isn't able to handle categoricals, one-hot-encode categoricals
        is_tree_model = explainable_model.explainable_model_type() == ExplainableModelType.TREE_EXPLAINABLE_MODEL_TYPE
        if is_tree_model and self._supports_categoricals(explainable_model):
            # Index the categorical string columns for training data
            self._column_indexer = initialization_examples.string_index(columns=categorical_features)
            self._one_hot_encoder = None
            if categorical_features:
                explainable_model_args[LightGBMParams.CATEGORICAL_FEATURE] = categorical_features
        else:
            # One-hot-encode categoricals for models that don't support categoricals natively
            self._column_indexer = initialization_examples.string_index(columns=categorical_features)
            self._one_hot_encoder = initialization_examples.one_hot_encode(columns=categorical_features)

        self.classes = classes
        self.explain_subset = explain_subset
        self.transformations = transformations
        self._shap_values_output = shap_values_output
        # Train the mimic model on the given model
        training_data = initialization_examples.dataset
        self.initialization_examples = initialization_examples
        if str(type(training_data)).endswith(".DenseData'>"):
            training_data = training_data.data

        explainable_model_args[ExplainParams.CLASSIFICATION] = self.predict_proba_flag
        if self._supports_shap_values_output(explainable_model):
            explainable_model_args[ExplainParams.SHAP_VALUES_OUTPUT] = shap_values_output
        self.surrogate_model = _model_distill(self.function, explainable_model, training_data,
                                              original_training_data, explainable_model_args)
        self._method = self.surrogate_model._method
        self._original_eval_examples = None
        self._allow_all_transformations = allow_all_transformations

    def _get_transformed_data(self, evaluation_examples):
        """Return the transformed data for some evaluation data.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which to
            explain the model's output.  If specified, computes feature importance through aggregation.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: Transformed data.
        :rtype: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        """
        if self.transformations is not None:
            _, transformed_evaluation_examples = get_datamapper_and_transformed_data(
                examples=evaluation_examples, transformations=self.transformations,
                allow_all_transformations=self._allow_all_transformations)
        else:
            transformed_evaluation_examples = evaluation_examples

        return transformed_evaluation_examples

    def _get_surrogate_model_predictions(self, evaluation_examples):
        """Return the predictions given by the surrogate model.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which to
            explain the model's output.  If specified, computes feature importance through aggregation.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: predictions of the surrogate model.
        :rtype: numpy.array
        """
        transformed_evaluation_examples = self._get_transformed_data(evaluation_examples)
        if self.classes is not None and len(self.classes) == 2:
            index_predictions = _inverse_soft_logit(self.surrogate_model.predict(transformed_evaluation_examples))
            actual_predictions = []
            for index in index_predictions:
                actual_predictions.append(self.classes[index])
            return np.array(actual_predictions)
        else:
            return self.surrogate_model.predict(transformed_evaluation_examples)

    def _get_teacher_model_predictions(self, evaluation_examples):
        """Return the predictions given by the teacher model.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which to
            explain the model's output.  If specified, computes feature importance through aggregation.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: predictions of the surrogate model.
        :rtype: numpy.array
        """
        transformed_evaluation_examples = self._get_transformed_data(evaluation_examples)
        return self.model.predict(transformed_evaluation_examples)

    def _supports_categoricals(self, explainable_model):
        return issubclass(explainable_model, LGBMExplainableModel)

    def _supports_shap_values_output(self, explainable_model):
        return issubclass(explainable_model, LGBMExplainableModel)

    def _get_explain_global_kwargs(self, evaluation_examples=None, include_local=True,
                                   batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Get the kwargs for explain_global to create a global explanation.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which to
            explain the model's output.  If specified, computes feature importance through aggregation.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param include_local: Include the local explanations in the returned global explanation.
            If evaluation examples are specified and include_local is False, will stream the local
            explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: Args for explain_global.
        :rtype: dict
        """
        classification = self.predict_proba_flag
        kwargs = {ExplainParams.METHOD: self._get_method}
        if classification:
            kwargs[ExplainParams.CLASSES] = self.classes
        if evaluation_examples is not None:

            # Aggregate local explanation to global, either through computing the local
            # explanation and then aggregating or streaming the local explanation to global
            if include_local:
                # Get local explanation
                local_explanation = self._explain_local(evaluation_examples, from_global=True)
                kwargs[ExplainParams.LOCAL_EXPLANATION] = local_explanation
            else:
                if classification:
                    model_task = ModelTask.Classification
                else:
                    model_task = ModelTask.Regression
                if not isinstance(evaluation_examples, DatasetWrapper):
                    self._logger.debug('Eval examples not wrapped, wrapping')
                    evaluation_examples = DatasetWrapper(evaluation_examples)

                kwargs = _aggregate_streamed_local_explanations(self, evaluation_examples, model_task, self.features,
                                                                batch_size, **kwargs)
            return kwargs
        global_importance_values = self.surrogate_model.explain_global()
        order = _order_imp(global_importance_values)
        if classification:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.CLASSIFICATION
        else:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.REGRESSION
        if self.model is not None:
            kwargs[ExplainParams.MODEL_TYPE] = str(type(self.model))
        else:
            kwargs[ExplainParams.MODEL_TYPE] = ExplainType.FUNCTION
        kwargs[ExplainParams.EXPECTED_VALUES] = None
        kwargs[ExplainParams.CLASSIFICATION] = classification
        kwargs[ExplainParams.GLOBAL_IMPORTANCE_VALUES] = global_importance_values
        kwargs[ExplainParams.GLOBAL_IMPORTANCE_RANK] = order
        kwargs[ExplainParams.FEATURES] = self.features
        return kwargs

    def explain_global(self, evaluation_examples=None, include_local=True,
                       batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Globally explains the blackbox model using the surrogate model.

        If evaluation_examples are unspecified, retrieves global feature importance from explainable
        surrogate model.  Note this will not include per class feature importance. If evaluation_examples
        are specified, aggregates local explanations to global from the given evaluation_examples - which
        computes both global and per class feature importance.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which to
            explain the model's output.  If specified, computes feature importance through aggregation.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param include_local: Include the local explanations in the returned global explanation.
            If evaluation examples are specified and include_local is False, will stream the local
            explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: A model explanation object. It is guaranteed to be a GlobalExplanation. If evaluation_examples are
            passed in, it will also have the properties of a LocalExplanation. If the model is a classifier (has
            predict_proba), it will have the properties of ClassesMixin, and if evaluation_examples were passed in it
            will also have the properties of PerClassMixin.
        :rtype: DynamicGlobalExplanation
        """
        if self._original_eval_examples is None:
            if isinstance(evaluation_examples, DatasetWrapper):
                self._original_eval_examples = evaluation_examples.original_dataset_with_type
            else:
                self._original_eval_examples = evaluation_examples
        kwargs = self._get_explain_global_kwargs(evaluation_examples=evaluation_examples, include_local=include_local,
                                                 batch_size=batch_size)
        kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
        has_eval_examples = evaluation_examples is not None
        if has_eval_examples:
            kwargs[ExplainParams.EVAL_DATA] = self._original_eval_examples
            ys_dict = self._get_ys_dict(self._original_eval_examples,
                                        transformations=self.transformations,
                                        allow_all_transformations=self._allow_all_transformations)
            kwargs.update(ys_dict)
            if include_local:
                return _aggregate_global_from_local_explanation(**kwargs)
        explanation = _create_global_explanation(**kwargs)
        has_data_mapper = self._datamapper is not None

        if has_data_mapper:
            if not has_eval_examples:
                # If transformations have been passed, and no evaluation data, then return raw features explanation
                raw_kwargs = _get_raw_explainer_create_explanation_kwargs(kwargs=kwargs)
                return _create_raw_feats_global_explanation(explanation,
                                                            feature_maps=[self._datamapper.feature_map],
                                                            features=self.features,
                                                            **raw_kwargs)
            else:
                # Note this is a raw explanation already if transformations were passed
                explanation._is_raw = True
        # If eval examples passed, even if transformations is passed the explanation is
        # already in terms of raw data since it is aggregated from engineered.
        # If there is no data mapper then we should return the explanation directly.
        return explanation

    @property
    def _get_method(self):
        """Get the method for this explainer, or mimic with surrogate model type.

        :return: The method, or mimic with surrogate model type.
        :rtype: str
        """
        return "{}.{}".format(ExplainType.MIMIC, self._method)

    def _get_explain_local_kwargs(self, evaluation_examples, from_global):
        """Get the kwargs for explain_local to create a local explanation.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param from_global: True if this is called from the explain_global API.
        :type from_global: bool
        :return: Args for explain_local.
        :rtype: dict
        """
        if self.reset_index != ResetIndex.Ignore:
            evaluation_examples.reset_index()
        kwargs = {}
        probabilities = None
        if self._shap_values_output == ShapValuesOutput.TEACHER_PROBABILITY:
            original_evaluation_examples = evaluation_examples.typed_dataset
            # Outputting shap values in terms of the probabilities of the teacher model
            probabilities = self.function(original_evaluation_examples)
        # if index column should not be set on surrogate model, remove it
        if self.reset_index == ResetIndex.ResetTeacher:
            evaluation_examples.set_index()
        if self._timestamp_featurizer:
            evaluation_examples.apply_timestamp_featurizer(self._timestamp_featurizer)
        if self._column_indexer:
            evaluation_examples.apply_indexer(self._column_indexer, bucket_unknown=True)
        if self._one_hot_encoder:
            evaluation_examples.apply_one_hot_encoder(self._one_hot_encoder)

        dataset = evaluation_examples.dataset

        kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features

        local_importance_values = self.surrogate_model.explain_local(dataset, probabilities=probabilities)
        classification = isinstance(local_importance_values, list) or self.predict_proba_flag
        is_sparse = issparse(local_importance_values) or issparse(local_importance_values[0])
        expected_values = self.surrogate_model.expected_values
        kwargs[ExplainParams.METHOD] = self._get_method
        self.features = evaluation_examples.get_features(features=self.features)
        kwargs[ExplainParams.FEATURES] = self.features

        if self.predict_proba_flag:
            if not is_sparse:
                if self.surrogate_model.multiclass:
                    # For multiclass case, convert to array, but only if not sparse
                    local_importance_values = np.array(local_importance_values)
                else:
                    # TODO: Eventually move this back inside the surrogate model
                    # If binary case, we need to reformat the data to have importances per class
                    # and convert the expected values back to the original domain
                    local_importance_values = np.stack((-local_importance_values, local_importance_values))
            elif not self.surrogate_model.multiclass:
                # For binary classification sparse case we need to reformat the data
                # to have importance values per class
                local_importance_values = [-local_importance_values, local_importance_values]

        if classification:
            kwargs[ExplainParams.CLASSES] = self.classes
        # Reformat local_importance_values result if explain_subset specified
        if self.explain_subset:
            self._logger.debug('Getting subset of local_importance_values')
            if classification:
                if is_sparse and self.surrogate_model.multiclass:
                    for i in range(len(local_importance_values)):
                        local_importance_values[i] = local_importance_values[i][:, self.explain_subset]
                else:
                    local_importance_values = local_importance_values[:, :, self.explain_subset]
            else:
                local_importance_values = local_importance_values[:, self.explain_subset]
        if classification:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.CLASSIFICATION
        else:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.REGRESSION
        if self.model is not None:
            kwargs[ExplainParams.MODEL_TYPE] = str(type(self.model))
        else:
            kwargs[ExplainParams.MODEL_TYPE] = ExplainType.FUNCTION
        kwargs[ExplainParams.LOCAL_IMPORTANCE_VALUES] = local_importance_values
        kwargs[ExplainParams.EXPECTED_VALUES] = np.array(expected_values)
        kwargs[ExplainParams.CLASSIFICATION] = classification
        if not from_global:
            ys_dict = self._get_ys_dict(self._original_eval_examples,
                                        transformations=self.transformations,
                                        allow_all_transformations=self._allow_all_transformations)
            kwargs.update(ys_dict)
        return kwargs

    def _explain_local_helper(self, evaluation_examples, from_global):
        """Helper method to create the local explanation.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param from_global: True if this is called from the explain_global API.
        :type from_global: bool
        :return: kwargs for building the model explanation object
        :rtype: dict
        """
        if not from_global:
            if self._original_eval_examples is None:
                if isinstance(evaluation_examples, DatasetWrapper):
                    self._original_eval_examples = evaluation_examples.original_dataset_with_type
                else:
                    self._original_eval_examples = evaluation_examples
        if self._datamapper is not None:
            new_evaluation_examples = transform_with_datamapper(evaluation_examples, self._datamapper)
            # memory optimization to clear all internal references so they can be gc'ed as soon as possible
            evaluation_examples._clear()
            evaluation_examples = new_evaluation_examples

        kwargs = self._get_explain_local_kwargs(evaluation_examples, from_global)
        if not from_global:
            kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
            kwargs[ExplainParams.EVAL_DATA] = evaluation_examples
        return kwargs

    @tabular_decorator
    def _explain_local(self, evaluation_examples, from_global=False):
        """Locally explains the blackbox model using the surrogate model.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param from_global: True if this is called from the explain_global API.
        :type from_global: bool
        :return: A model explanation object. It is guaranteed to be a LocalExplanation. If the model is a classifier,
            it will have the properties of the ClassesMixin.
        :rtype: DynamicLocalExplanation
        """
        kwargs = self._explain_local_helper(evaluation_examples, from_global)
        explanation = _create_local_explanation(**kwargs)

        # if transformations have been passed, then return raw features explanation
        raw_kwargs = _get_raw_explainer_create_explanation_kwargs(kwargs=kwargs)

        return explanation if self._datamapper is None else _create_raw_feats_local_explanation(
            explanation, feature_maps=[self._datamapper.feature_map], features=self.features, **raw_kwargs)

    @tabular_decorator
    def explain_local(self, evaluation_examples):
        """Locally explains the blackbox model using the surrogate model.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: A model explanation object. It is guaranteed to be a LocalExplanation. If the model is a classifier,
            it will have the properties of the ClassesMixin.
        :rtype: DynamicLocalExplanation
        """
        return self._explain_local(evaluation_examples)

    def _save(self):
        """Return a string dictionary representation of the mimic explainer.

        Currently only supported scenario is Mimic Explainer with LightGBM surrogate model.

        :return: A serialized dictionary representation of the mimic explainer.
        :rtype: dict
        """
        properties = {}
        # save all of the properties
        for key, value in self.__dict__.items():
            if key in MimicSerializationConstants.nonify_properties:
                properties[key] = None
            elif key in MimicSerializationConstants.save_properties:
                properties[key] = value._save()
            else:
                properties[key] = json.dumps(value)
        # return a dictionary of strings
        return properties

    @staticmethod
    def _load(model, properties):
        """Load a MimicExplainer from the given properties.

        Currently only supported scenario is Mimic Explainer with LightGBM surrogate model.

        :param model: The serialized ONNX model with a scikit-learn like API.
        :type model: ONNX model.
        :param properties: A serialized dictionary representation of the mimic explainer.
        :type properties: dict
        :return: The deserialized MimicExplainer.
        :rtype: interpret_community.mimic.MimicExplainer
        """
        # create the MimicExplainer without any properties using the __new__ function, similar to pickle
        mimic = MimicExplainer.__new__(MimicExplainer)
        # load all of the properties
        for key, value in properties.items():
            # Regenerate the properties on the fly
            if key in MimicSerializationConstants.nonify_properties:
                if key == MimicSerializationConstants.MODEL:
                    mimic.__dict__[key] = model
                elif key == MimicSerializationConstants.LOGGER:
                    parent = logging.getLogger(__name__)
                    mimic_identity = json.loads(properties[MimicSerializationConstants.IDENTITY])
                    mimic.__dict__[key] = parent.getChild(mimic_identity)
                elif key == MimicSerializationConstants.INITIALIZATION_EXAMPLES:
                    mimic.__dict__[key] = None
                elif key == MimicSerializationConstants.ORIGINAL_EVAL_EXAMPLES:
                    mimic.__dict__[key] = None
                elif key == MimicSerializationConstants.TIMESTAMP_FEATURIZER:
                    mimic.__dict__[key] = None
                elif key == MimicSerializationConstants.FUNCTION:
                    # TODO add third case if is_function was passed to mimic explainer
                    if json.loads(properties[MimicSerializationConstants.PREDICT_PROBA_FLAG]):
                        mimic.__dict__[key] = model.predict_proba
                    else:
                        mimic.__dict__[key] = model.predict
                else:
                    raise Exception("Unknown nonify key on deserialize in MimicExplainer: {}".format(key))
            elif key in MimicSerializationConstants.save_properties:
                mimic.__dict__[key] = LGBMExplainableModel._load(value)
            elif key in MimicSerializationConstants.enum_properties:
                # NOTE: If more enums added in future, will need to handle this differently
                mimic.__dict__[key] = ShapValuesOutput(json.loads(value))
            else:
                mimic.__dict__[key] = json.loads(value)
        if MimicSerializationConstants.ORIGINAL_EVAL_EXAMPLES not in mimic.__dict__:
            mimic.__dict__[MimicSerializationConstants.ORIGINAL_EVAL_EXAMPLES] = None
        if MimicSerializationConstants.TIMESTAMP_FEATURIZER not in mimic.__dict__:
            mimic.__dict__[MimicSerializationConstants.TIMESTAMP_FEATURIZER] = None
        if MimicSerializationConstants.RESET_INDEX not in mimic.__dict__:
            mimic.__dict__[MimicSerializationConstants.RESET_INDEX] = False
        if MimicSerializationConstants.ALLOW_ALL_TRANSFORMATIONS not in mimic.__dict__:
            mimic.__dict__[MimicSerializationConstants.ALLOW_ALL_TRANSFORMATIONS] = False
        return mimic

    def __getstate__(self):
        """Influence how MimicExplainer is pickled.

        Removes logger which is not serializable.

        :return state: The state to be pickled, with logger removed.
        :rtype: dict
        """
        odict = self.__dict__.copy()
        del odict['_logger']
        return odict

    def __setstate__(self, state):
        """Influence how MimicExplainer is unpickled.

        Re-adds logger which is not serializable.

        :param dict: A dictionary of deserialized state.
        :type dict: dict
        """
        self.__dict__.update(state)
        self._logger = logging.getLogger(__name__)

    def _get_surrogate_model_replication_measure(self, training_data):
        """Return the metric which tells how well the surrogate model replicates the teacher model.
        :param training_data: The data for getting the replication metric.
        :type training_data: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: Metric that tells how well the surrogate model replicates the behavior of teacher model.
        :rtype: float
        """
        if self.classes is None and training_data.shape[0] == 1:
            raise ScenarioNotSupportedException(
                "Replication measure for regression surrogate not supported "
                "because of single instance in training data")

        surrogate_model_predictions = self._get_surrogate_model_predictions(training_data)
        teacher_model_predictions = self._get_teacher_model_predictions(training_data)

        if self.classes is not None:
            replication_measure = accuracy_score(teacher_model_predictions, surrogate_model_predictions)
        else:
            replication_measure = r2_score(teacher_model_predictions, surrogate_model_predictions)
        return replication_measure
